import numpy as np
import math
import warnings
from sklearn import linear_model
from scipy.optimize import curve_fit
import gradientDescent as GD
warnings.filterwarnings("error")

def ErrorGenerator (sigmaRoot):
    Error = []
    for i in range (len(sigmaRoot)):
        s = np.random.normal(0, 1, None)
        Error.append(s)
    Error = np.matrix([Error])
    Error = Error.transpose()
    Error = sigmaRoot*Error
    return Error

def YGenerator (x1, Atrue, Ctrue, Qtrue, Rtrue, P, length):
    x = np.matrix(x1)
    x = x.transpose() #column
    V, S, U = np.linalg.svd(P, full_matrices = True)
    Ssqrt = np.zeros(shape = (len(S), len(S)))
    for i in range(len(S)):
        Ssqrt[i, i] = np.sqrt(S[i])
    rootSigma = V*Ssqrt*np.transpose(V)
    x = x + ErrorGenerator(rootSigma)
    print x
    Y = [] #Y will include the whole series of observations
    for i in range(length):#creating the series of "length" length observations
        y = Ctrue*x + ErrorGenerator(Rtrue) #create observation
        x = Atrue*x + ErrorGenerator(Qtrue) #update states
        Y.append(y) #add to the Y
    return Y

###########DEBUGGED###############
#this generate the observation sequence according to the PLG model
def generate(mu, Sigma, G, sigma, C, m, length):
    #here the sigma is of dimension m by m insteadl of m*n by m*n and G is of dimension, C here is dimension m*n by m and G is of dimension  
    Ys = []
    Sigmas = []
    mus = []
    #print Sigma
    for j in range(length):
        Sigmas.append(Sigma[0:m, 0:m])
        mus.append(mu[0:m, :])
        #print Sigma[0:m, 0:m]
        # taking the root of sigma[1, 1]
        V, S, U = np.linalg.svd(Sigma[0:m, 0:m], full_matrices = True)
        #V, S, U = np.linalg.svd(Sigma, full_matrices = True)
        Ssqrt = np.zeros(shape = (len(S), len(S)))
        for i in range(len(S)):
            Ssqrt[i, i] = np.sqrt(S[i])
        rootSigma = V*Ssqrt*np.transpose(V)
        
        #producing an observation
        e = ErrorGenerator(rootSigma)
        y = np.matrix(mu)[0:m] + e
        #y = (np.matrix(mu) + e)[0:m]
        Ys.append(y)
        mu_temp = np.concatenate((mu, G*mu), axis=0)
        
        #calculating the joint distribution of the mu and the new observation generated
        left = np.concatenate((Sigma, G*Sigma+np.transpose(C)), axis = 0)
        right = np.concatenate((Sigma*np.transpose(G)+C, G*Sigma*np.transpose(G)+sigma+G*C+np.transpose(C)*np.transpose(G)), axis = 0)
        sigma_temp = np.concatenate((left, right), axis = 1)
        
        #update the mu and Sigma
        mu = mu_temp[m:]+sigma_temp[m:, 0:m]*np.linalg.solve(Sigma[0:m, 0:m], np.identity(m))*e
        Sigma = sigma_temp[m:, m:] - sigma_temp[m:, 0:m]*np.linalg.solve(Sigma[0:m, 0:m], np.identity(m))*sigma_temp[0:m, m:]
    return Ys, Sigmas, mus
###################################


def likelihood(Sigmas, Mus, Ys, k, printT):
    sum1 = 0
    sum2 = 0
    done = 0
    #result = 1
    #print Sigmas
    try:
        for i in range(len(Sigmas)):
            if (np.transpose(Ys[i]-Mus[i])*np.linalg.pinv(Sigmas[i])*(Ys[i]-Mus[i]))[0, 0] <= 0 or np.linalg.det(Sigmas[i])<= 0:
                return 0
            sum1 = sum1 + math.log(np.linalg.det(Sigmas[i]))
            sum2 = sum2 + (np.transpose(Ys[i]-Mus[i])*np.linalg.pinv(Sigmas[i])*(Ys[i]-Mus[i]))[0,0]
            #print "second time ", i
            if i < 3:
                """print "new step"
                print Ys[i]
                print Mus[i]
                print np.linalg.pinv(Sigmas[i])
                print np.linalg.det(Sigmas[i])
                print Sigmas[i]
                print (np.transpose(Ys[i]-Mus[i])*np.linalg.pinv(Sigmas[i])*(Ys[i]-Mus[i]))[0,0]
                probability = ((2*math.pi)**(-k/2))*(np.linalg.det(Sigmas[i])**(-1/2))*(np.e**((-1*np.transpose(Ys[i]-Mus[i])*np.linalg.pinv(Sigmas[i])*(Ys[i]-Mus[i])/2)[0, 0]))
                print probability"""
        result = -float(k)*float(len(Ys))*math.log(2.0*math.pi)/2.0 - 0.5*(sum1+sum2)
        """if result > 0:
            print "new term"
            print sum1
            print sum2
            print result"""
        #if printT == True:
        """print "TOTAL"
        print sum1
        print sum2
        print -float(k)*float(len(Ys))/2.0*math.log(2.0*math.pi)
        print result"""
        #print result
        #print 0.5*(sum1 + sum2)
        #print -k*len(Ys)/2*math.log(2*math.pi)
    except ValueError:
        return 0
    if done == 1:
        return 0
    return result
    
def func1(state, a, b, c, d):
    return a*state[0] + b*state[1] + c*state[2] + d*state[3]

def func2(state, a, b, c, d):
    return a*state[0] + b*state[1] + c*state[2] + d*state[3]

###################DEBUGGED######################
def learning(Ymatrices, Os, m, n, method, reduced):
    """ 
    method 0 is covariance method with SVD
    method 1 is least square method without averaging
    method 2 is least square method with averaging
    method 3 is covariance method without SVD
    method 4 is least square method with enforced sparsity 
    method 5 is for least square used in real data
    method 6 is least square with pre-imposed sparsity pattern
    method 7 is gradient descent 
    """
    # estimate matrix O, H and mu by taking the average
    if method == 1 or method == 2 or method == 4 or method == 6:
        O = Os[2][0]
        averageH = Ymatrices[0]
    mu = np.transpose(np.matrix(Ymatrices[0][:, 0]))
    
    for i in range(1,len(Ymatrices)):
        
        mu = mu + np.transpose(np.matrix(Ymatrices[i][:, 0]))
        if method == 1 or method == 4 or method == 6:
            averageH = np.concatenate((averageH,Ymatrices[i]), axis = 1)
            O = np.concatenate((O, Os[2][i]), axis = 1)
        elif method == 2:
            averageH = averageH + Ymatrices[i]
            O = O + Os[2][i]
    if method == 2:
        averageH = averageH/len(Ymatrices)
        O = O/len(Os[2])
    mu = mu / len(Ymatrices)
    G = np.zeros((m, n*m))
    """if method == 6:
        fH = []
        sH = []
        tH = []
        for i in range(averageH.shape[1]):
            fH.append([averageH[0][i], averageH[3][i], averageH[6][i], averageH[9][i], averageH[12][i], averageH[15][i]])
            sH.append([averageH[1][i], averageH[4][i], averageH[7][i], averageH[10][i], averageH[13][i], averageH[16][i]])
            tH.append([averageH[2][i], averageH[5][i], averageH[8][i], averageH[11][i], averageH[14][i], averageH[17][i]])
        fH = np.transpose(np.matrix(fH))
        sH = np.transpose(np.matrix(sH))
        tH = np.transpose(np.matrix(tH))
        print averageH.shape        
        print fH.shape
        print O.shape
        popt0 = O[0, :]*np.transpose(fH)*np.linalg.pinv(fH*np.transpose(fH))
        popt1 = O[1, :]*np.transpose(sH)*np.linalg.pinv(sH*np.transpose(sH))
        popt2 = O[2, :]*np.transpose(tH)*np.linalg.pinv(tH*np.transpose(tH))
        print popt0.shape
        G = np.matrix([[popt0[0,0], 0.0, 0.0, popt0[0,1], 0.0, 0.0, popt0[0,2], 0.0, 0.0, popt0[0,3], 0.0, 0.0, popt0[0, 4], 0.0, 0.0, popt0[0, 5], 0.0, 0.0], [0.0, popt1[0,0], 0.0, 0.0, popt1[0,1], 0.0, 0.0, popt1[0,2], 0.0, 0.0, popt1[0,3], 0.0, 0.0, popt1[0, 4], 0.0, 0.0, popt1[0, 5], 0.0], [0.0, 0.0, popt2[0,0], 0.0, 0.0, popt2[0,1], 0.0, 0.0, popt2[0,2], 0.0, 0.0, popt2[0,3], 0.0, 0.0, popt2[0, 4], 0.0, 0.0, popt2[0, 5]]])"""
    #estimate G
    if method == 6:
        G = []
        for i in range(reduced.shape[0]):
            info = []
            pattern = []
            for k in range(reduced.shape[1]):
                if abs(reduced[i,k]) > 0.001:
                    pattern.append(k)
            for s in range(averageH.shape[1]):
                col = []
                for item in pattern:
                    col.append(averageH[item][s])
                info.append(col)
            H = np.transpose(np.matrix(info))
            popt = O[i, :]*np.transpose(H)*np.linalg.pinv(H*np.transpose(H))
            Grow = []
            for j in range(reduced.shape[1]):
                Grow.append(0.0)
            incre = 0
            for item in pattern:
                Grow[item] = popt[0, incre]
                incre = incre+1
            G.append(Grow)
        G = np.matrix(G)
        print G
    if method == 1 or method == 2:
        squareH = np.dot(averageH,np.transpose(averageH))
        G = O * np.transpose(averageH)*np.linalg.pinv(squareH)
    if method == 5:
        squareH = np.dot(Os[3],np.transpose(Os[3]))
        G = Os[4] * np.transpose(Os[3])*np.linalg.pinv(squareH)
    if method == 0 or method == 7:
        #print Os[1]
        try:
            #print Os[1].shape
            U, s, Vt = np.linalg.svd(Os[1])
            #print s
            #u, s, vt = svds(Os[1], reduced)
            d = np.eye(reduced)
            for i in range(reduced):
                d[i][i] = 1/s[i]
            G = Os[0]*np.transpose(Vt[0:reduced, :])*d*np.transpose(U[:, 0:reduced])
        except RuntimeWarning and np.linalg.linalg.LinAlgError:
            print "SVD FAILS"
            #print s
            #print Os[1]
            return (np.inf, np.inf, np.inf, np.inf, np.inf)
    if method== 4:
        G = []
        #print averageH.shape
        #print (O[0, :].flatten()).shape
        
        clf = linear_model.Lasso(alpha=reduced, fit_intercept = False, max_iter = 10000000)
        #clf.get_params()
        for i in range(m):
            tempO = []
            for j in range((O[i, :]).shape[1]):
                tempO.append(O[i, j])
            #print O[i, 0:10]
            clf.fit(np.transpose(averageH), tempO)
            G.append(clf.coef_)
            #print clf.coef_
        G = np.matrix(G)
        
    if method == 3:
        try:
            G = Os[0]*np.linalg.pinv(Os[1])       
        except RuntimeWarning and np.linalg.linalg.LinAlgError:
            return (np.inf, np.inf, np.inf, np.inf, np.inf)
    #estimate the initial state covariance 
    Sigma =(Ymatrices[0][:, 0] - mu)*np.transpose(Ymatrices[0][:, 0] - mu)
    for i in range(1, len(Ymatrices)):
        Sigma = Sigma + (np.transpose(np.matrix(Ymatrices[i][:, 0])) - mu)*np.transpose(np.transpose(np.matrix(Ymatrices[i][:, 0])) - mu)
    Sigma = Sigma/len(Ymatrices)

    #construct the Gprime matrix for updating the state vector
    zero = np.zeros(((n-1)*m, m))
    top = np.concatenate((zero,np.identity((n-1)*m)), axis = 1)
    Gprime = np.concatenate((top, G), axis = 0)
    
    #Learning the covariance and noise covariance matrix
    CT = np.zeros((n*m, n*m))
    SigmaDeltaT = np.zeros((n*m, n*m))
    SigmaDelta = np.zeros((m, m))
    C = np.zeros((n*m, m))
    count = 0
    for i in range(len(Ymatrices)):
        for t in range(Ymatrices[i].shape[1]):
            delta = Os[2][i][:, t]-G*np.transpose(np.matrix(Ymatrices[i][:, t]))
            SigmaDelta = SigmaDelta + delta*np.transpose(delta)
            C = C + np.transpose(np.matrix(Ymatrices[i][:, t]))*np.transpose(delta)
            count = count + 1
    C = C/count
    CT[:, (n-1)*m:] = C
    SigmaDelta = SigmaDelta/count
    SigmaDeltaT[(n-1)*m:, (n-1)*m:] = SigmaDelta   
    return (mu, Gprime, CT, SigmaDeltaT, Sigma)
#################################################

#predict a new observation according to the past observation
def prediction(mu, Gprime, C, SigmaDelta, Sigma, m, n, y, initSigma):
    J = np.matrix(np.concatenate((np.identity(m), np.zeros((m, (n-1)*m))), axis=1))
    #print m
    F = (Gprime*Sigma + np.transpose(C))*np.transpose(J)
    #print "pre inverse"
    #print J*Sigma*np.transpose(J)
    #U, d, V = np.linalg.svd(J*Sigma*np.transpose(J))
    """print "svd"
    print d
    print "vectors"
    print U"""
    s = np.linalg.pinv(J*Sigma*np.transpose(J))
    #print "post inverse"    
    #print s
    """print mu.shape
    print Gprime.shape
    print F.shape
    print s.shape
    print y.shape
    print y
    print (J*mu).shape"""
    mu = Gprime * mu + F*s*(y-J*mu)
     
    GG = Gprime*Sigma*np.transpose(Gprime)
    GC = Gprime*C
    CG = np.transpose(C)*np.transpose(Gprime)
    try:
        Sigma =  GG + GC + CG + SigmaDelta-F*s*np.transpose(F)
        #Sigma =  GG + GC + CG + SigmaDelta
    except RuntimeWarning:
        #print "diverge"
        Sigma = initSigma
    if math.isnan(Sigma[0, 0]):
        #print "diverge"
        Sigma = initSigma
    return (mu, Sigma)
